/* MoreBalance - load balancer and networking tool
 * Copyright (C) 2002 - 2006 Josef Spillner <josef@coolprojects.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* MoreBalance includes */
#include "magic.h"

/* System includes */
#include <stdlib.h>
#include <stdio.h> /**/
#include <string.h>
#include <time.h>

/* Global variables */
struct magic_t
{
	int *ports;
	int *destports;
	int numports;
	char ***nodes;
	int *numnodes;
	int *curnodes;
	long long **traffic;
};

struct hash_t
{
	int *peers;
	int *ports;
	int *stamps;
	char **nodes;
	int numpeers;
};

typedef struct magic_t magic;
typedef struct hash_t hash;

static magic m;
static hash h;

/* Initialize all the magic */
void magic_initialize(conf *c)
{
	int i, j, k, pnumber, cnumber;

	m.ports = NULL;
	m.destports = NULL;
	m.nodes = NULL;
	m.numnodes = NULL;
	m.curnodes = NULL;
	m.traffic = NULL;
	m.numports = 0;

	h.peers = NULL;
	h.ports = NULL;
	h.stamps = NULL;

	for(i = 0; i < c->clusternum; i++)
	{
		for(j = 0; j < c->cluster[i]->portsnum; j++)
		{
			/* Find port-based list in magic */
			pnumber = -1;
			for(k = 0; k < m.numports; k++)
			{
				if(m.ports[k] == c->cluster[i]->ports[j])
				{
					pnumber = k;
					break;
				}
			}

			if(pnumber < 0)
			{
				/* New port-based list to be created */
				pnumber = m.numports;
				m.ports = (int*)realloc(m.ports, (pnumber + 1) * sizeof(int));
				m.ports[pnumber] = c->cluster[i]->ports[j];
				m.destports = (int*)realloc(m.destports, (pnumber + 1) * sizeof(int));
				m.destports[pnumber] = c->cluster[i]->destport;

				m.curnodes = (int*)realloc(m.curnodes, (pnumber + 1) * sizeof(int));
				m.curnodes[pnumber] = 0;

				m.numnodes = (int*)realloc(m.numnodes, (pnumber + 1) * sizeof(int));
				m.numnodes[pnumber] = 0;

				m.nodes = (char***)realloc(m.nodes, (pnumber + 1) * sizeof(char**));
				m.nodes[pnumber] = NULL;

				m.traffic = (long long**)realloc(m.traffic, (pnumber + 1) * sizeof(long long*));
				m.traffic[pnumber] = NULL;

				m.numports++;
			}

			/* Fill port-based list with nodes */
			cnumber = m.numnodes[pnumber];
			for(k = 0; k < c->cluster[i]->nodesnum; k++)
			{
				m.nodes[pnumber] = (char**)realloc(m.nodes[pnumber], (cnumber + 1) * sizeof(char*));
				m.nodes[pnumber][cnumber] = strdup(c->cluster[i]->nodes[k]);

				m.traffic[pnumber] = (long long*)realloc(m.traffic[pnumber], (cnumber + 1) * sizeof(long long));
				m.traffic[pnumber][cnumber] = 0;

				cnumber++;
			}
			m.numnodes[pnumber] = cnumber;
		}
	}
}

/* Disable one node */
void magic_disable(const char *node, int port)
{
	int i, k;

	for(i = 0; i < m.numports; i++)
	{
		if(m.destports[i] == port)
		{
			for(k = 0; k < m.numnodes[i]; k++)
			{
				if(!strcmp(node, m.nodes[i][k]))
				{
						printf("Disabling: %s for %i\n", node, port);
						free(m.nodes[i][k]);
						m.nodes[i][k] = strdup("");
				}
			}
		}
	}
}

/* Returns position in hash table, or -1 if not found */
int magic_hashget(int peer)
{
	int i;

	for(i = 0; i < h.numpeers; i++)
	{
		if(h.peers[i] == peer)
		{
			if(time(NULL) < h.stamps[i]) return i;
		}
	}

	return -1;
}

/* Add an entry to the hash table */
void magic_hashset(int peer, int port, const char *ret, int timeout)
{
	h.peers = (int*)realloc(h.peers, (h.numpeers + 1) * sizeof(int));
	h.ports = (int*)realloc(h.ports, (h.numpeers + 1) * sizeof(int));
	h.stamps = (int*)realloc(h.stamps, (h.numpeers + 1) * sizeof(int));
	h.nodes = (char**)realloc(h.nodes, (h.numpeers + 1) * sizeof(char*));
	h.peers[h.numpeers] = peer;
	h.ports[h.numpeers] = port;
	h.stamps[h.numpeers] = time(NULL) + timeout;
	h.nodes[h.numpeers] = strdup(ret);
	h.numpeers++;
}

/* Return a node depending on the ip and the port */
char *magic_node(conf *c, const char *ip, int port, int peer)
{
	int i, j, pnumber;
	char *ret;

	if((!m.ports) || (!m.nodes)) return NULL;

	/* Find position for this port in list */
	ret = NULL;
	i = 0;
	pnumber = -1;
	for(i = 0; i < m.numports; i++)
	{
		if(m.ports[i] == port)
		{
			pnumber = i;
			break;
		}
	}
	if(pnumber < 0) return NULL;

	/* Find possible QoS guarantee */
	for(i = 0; i < c->qosnum; i++)
	{
		if(port == c->qos[i]->port1)
		{
			if(c->verbose) printf("QoS: Require %i [port %i] for %i\n", i, port, peer);
			j = magic_hashget(peer);
			if(j >= 0)
			{
				ret = h.nodes[j];
				if(c->verbose) printf("QoS: Selected %s\n", ret);
			}
		}
	}

	/* Pattern matching */
	if(!ret)
	{
		for(i = 0; i < c->clusternum; i++)
		{
			if(c->cluster[i]->pattern) printf("!!! disregard required pattern %s !!!\n", c->cluster[i]->pattern);
		}
	}

	/* Round Robin */
	if(!ret)
	{
		if(m.nodes[pnumber][m.curnodes[pnumber]])
		{
			m.curnodes[pnumber]++;
			if(m.curnodes[pnumber] >= m.numnodes[pnumber]) m.curnodes[pnumber] = 0;
			ret = m.nodes[pnumber][m.curnodes[pnumber]];
		}
	}

	/* Find possible QoS prerequisite */
	for(i = 0; i < c->qosnum; i++)
	{
		if(port == c->qos[i]->port2)
		{
			if(c->verbose) printf("QoS: %s inserted at %i [port %i] for %i\n", ret, h.numpeers, port, peer);
			magic_hashset(peer, c->qos[i]->port1, ret, c->qos[i]->timeout);
		}
	}

	if(!strcmp(ret, ""))
	{
		return NULL;
	}

	return ret;
}

/* Return a destination socket in case there's no port */
char *magic_socket(conf *c, const char *ip, int port)
{
	int i;
	for(i = 0; i < c->clusternum; i++)
	{
		if(c->cluster[i]->ports[0] == port)
		{
			return c->cluster[i]->destsocket;
		}
	}
	return NULL;
}

/* Return a destination port depending on the ip and the port */
int magic_port(conf *c, const char *ip, int port)
{
	int i;
	for(i = 0; i < c->clusternum; i++)
	{
		if(c->cluster[i]->ports[0] == port)
		{
			if(!c->cluster[i]->destport)
			{
				if(!c->cluster[i]->destsocket) return port;
				else return 0;
			}
			return c->cluster[i]->destport;
		}
	}
	return port;
}

/* Update the traffic counter */
void magic_traffic(conf *c, const char *hostname, long long sbytes, long long rbytes)
{
	int i, j;

	for(i = 0; i < m.numports; i++)
	{
		for(j = 0; j < m.numnodes[i]; j++)
		{
			if(!strcmp(m.nodes[i][j], hostname))
			{
				m.traffic[i][j] += sbytes;
				m.traffic[i][j] += rbytes;
				if(c->verbose) printf("*** Total traffic for %s: %qi bytes\n", hostname, m.traffic[i][j]);
				return;
			}
		}
	}
}


